meta = """[MPA4A] 535
range=8128
periods=2
fstchan=0
holdafter=6e+006
sweepmode=35c8088
swpreset=65535
prena=4
syncout=8
fdac=0
rxdelay=0
cycles=2048
sequences=10
tagbits=15
dac0=21fc9
dac1=a037
dac2=a037
dac3=2037
dac4=2037
dac5=2037
rtpreset=7253.886
digio=600
digval=0
autoinc=1
savedata=3
mpafmt=asc
sephead=0
fmt=asc
smoothpts=5
[CHN1]
REPORT-FILE from 07/19/2017 06:15:58.539  written 07/19/2017 06:15:58
cmline0=07/19/2017 06:15:58.539
cmline1=A1
range=8128
active=1
bitshift=0
cftfak=2580100
evpreset=10
roimin=0
roimax=8100
caloff=0.000000
calfact=0.800000
calfact2=0
calfact3=0
calunit=nsec
caluse=1
[CHN2]
REPORT-FILE from 07/19/2017 06:15:58.539  written 07/19/2017 06:15:58
cmline0=07/19/2017 06:15:58.539
cmline1=A2
range=8128
active=1
bitshift=0
cftfak=2580100
evpreset=10
roimin=0
roimax=8100
caloff=0.000000
calfact=0.800000
calfact2=0
calfact3=0
calunit=nsec
caluse=1
[CHN3]
REPORT-FILE from 07/19/2017 06:15:58.539  written 07/19/2017 06:15:58
cmline0=07/19/2017 06:15:58.539
cmline1=A3
range=8128
active=1
bitshift=0
cftfak=2580100
evpreset=10
roimin=0
roimax=8100
caloff=0.000000
calfact=0.800000
calfact2=0
calfact3=0
calunit=nsec
caluse=1
;MPA4 #535
time_patch=5b
;datalength=8 bytes
;bit0..2: channel# 1..6 ( 3 bit)
;bit3: edge 0=up / 1=dn ( 1 bit)
;bit4 ..31: timedata    (28 bit), max sweep length 0.216 s
;bit32..47: sweeps      (16 bit)
;bit48..62: tag0..tag14 (15 bit)
;bit63:     data_lost   ( 1 bit)
[DATA]"""

meta_bytes = b"""[MPA4A] 535
range=8128
periods=2
fstchan=0
holdafter=6e+006
sweepmode=35c8088
swpreset=65535
prena=4
syncout=8
fdac=0
rxdelay=0
cycles=2048
sequences=10
tagbits=15
dac0=21fc9
dac1=a037
dac2=a037
dac3=2037
dac4=2037
dac5=2037
rtpreset=7253.886
digio=600
digval=0
autoinc=1
savedata=3
mpafmt=asc
sephead=0
fmt=asc
smoothpts=5
[CHN1]
REPORT-FILE from 07/19/2017 06:15:58.539  written 07/19/2017 06:15:58
cmline0=07/19/2017 06:15:58.539
cmline1=A1
range=8128
active=1
bitshift=0
cftfak=2580100
evpreset=10
roimin=0
roimax=8100
caloff=0.000000
calfact=0.800000
calfact2=0
calfact3=0
calunit=nsec
caluse=1
[CHN2]
REPORT-FILE from 07/19/2017 06:15:58.539  written 07/19/2017 06:15:58
cmline0=07/19/2017 06:15:58.539
cmline1=A2
range=8128
active=1
bitshift=0
cftfak=2580100
evpreset=10
roimin=0
roimax=8100
caloff=0.000000
calfact=0.800000
calfact2=0
calfact3=0
calunit=nsec
caluse=1
[CHN3]
REPORT-FILE from 07/19/2017 06:15:58.539  written 07/19/2017 06:15:58
cmline0=07/19/2017 06:15:58.539
cmline1=A3
range=8128
active=1
bitshift=0
cftfak=2580100
evpreset=10
roimin=0
roimax=8100
caloff=0.000000
calfact=0.800000
calfact2=0
calfact3=0
calunit=nsec
caluse=1
;MPA4 #535
time_patch=5b
;datalength=8 bytes
;bit0..2: channel# 1..6 ( 3 bit)
;bit3: edge 0=up / 1=dn ( 1 bit)
;bit4 ..31: timedata    (28 bit), max sweep length 0.216 s
;bit32..47: sweeps      (16 bit)
;bit48..62: tag0..tag14 (15 bit)
;bit63:     data_lost   ( 1 bit)
[DATA]"""



from typing import Dict, Tuple
import pandas as pd
import numpy as np
import attr
from itertools import chain


@attr.s(slots=True)
class MScanLineValidator:
    sig_val = attr.ib()  # SignalValidator

    def __getattr__(self, item):
        return getattr(self.sig_val, item)

    def run(self) -> Tuple[Dict, np.uint64]:
        """
        Interpolate MScan-specific line signals
        :return: Dictionary containing the data and the mean difference between subsequent lines
        """
        lines = self.dict_of_data["Lines"].loc[:, "abs_time"].copy()
        rel_idx, delta = self.__calc_line_parameters(lines=lines)
        lines, rel_idx, delta = self.__filter_extra_lines(lines=lines, delta=delta)
        if len(rel_idx) > 0:  # missing lines, not just extra
            theo_lines = self.__gen_line_model(lines=lines, m=delta)
            lines = self.__diff_vec_analysis(lines=lines, y=theo_lines, delta=delta)
        lines = self.__finalize_lines(lines=lines, delta=delta)
        if self.bidir:
            lines = self.sig_val.add_phase_to_bidir_lines(lines=lines)

        self.dict_of_data["Lines"] = pd.DataFrame(
            lines, dtype=np.uint64, columns=["abs_time"]
        )
        return self.dict_of_data, delta

    def __calc_line_parameters(self, lines: pd.Series) -> Tuple[np.ndarray, np.uint64]:
        """ Generate general parameters of the given acquisition """
        rel_idx = np.where(
            np.abs(lines.diff().pct_change(periods=1)) > self.change_thresh
        )[0]
        delta = np.uint64(
            lines.drop(rel_idx)
            .reindex(np.arange(len(lines)))
            .interpolate()
            .diff()
            .mean()
        )
        return rel_idx[::2], delta

    def __filter_extra_lines(
        self, lines: pd.Series, delta: np.uint64
    ) -> Tuple[pd.Series, pd.Series, np.uint64]:
        """
        Kick out excess line signals
        :param lines:
        :param delta:
        :return: Tuple of valid lines, missing lines and new delta of lines
        """
        diffs = lines.diff()
        rel_idx = np.where(np.abs(diffs.pct_change(periods=1)) > self.change_thresh)[0]
        recurring = np.where(np.diff(rel_idx) == 1)[0]
        idx_to_keep = np.ones_like(rel_idx, dtype=bool)

        for idx, _ in enumerate(recurring[1:], 1):
            try:
                if recurring[idx] - recurring[idx - 1] == 1:
                    idx_to_keep[recurring[idx] + 1] = False
            except IndexError:
                pass

        rel_idx_new = pd.Series(rel_idx[idx_to_keep][::2], dtype=np.uint64)
        missing_lines = []
        extra_lines = []

        for idx in rel_idx_new:
            if diffs[idx] < (delta / 2):  # excess lines
                extra_lines.append(idx)
            else:
                missing_lines.append(idx)

        valid_lines = lines.drop(extra_lines).reset_index(drop=True)
        delta = np.uint64(valid_lines.drop(missing_lines).diff().mean())

        # Get rid of lines that came after the last frame
        num_of_extra_lines = len(valid_lines) % self.num_of_lines
        valid_lines = valid_lines[:-num_of_extra_lines]

        return valid_lines, pd.Series(missing_lines), delta

    def __gen_line_model(self, lines: pd.Series, m: np.uint64) -> np.ndarray:
        """ Using linear approximation generate a model for the "correct" line signal """
        const = lines.iloc[0]
        x = np.arange(start=0, stop=len(lines), dtype=np.uint64)
        y = m * x + const
        if len(lines) > 1500:  # correct simulated lines
            idx_range = np.arange(1500, len(lines), step=1500, dtype=np.uint64)
            for idx in idx_range:
                x = np.arange(0, len(lines) - idx, dtype=np.uint64)
                y[idx:] = m * x + lines.iloc[idx]

        # MScan's lines are evenly separated
        first_diff = np.uint64(lines.iloc[2:110].diff()[1::2].median())
        delta_diff = np.int32((m - first_diff) / 2)
        if first_diff < m:
            y[1::2] -= delta_diff
            y[::2] += delta_diff
        else:
            y[1::2] += delta_diff
            y[::2] -= delta_diff
        return y

    def __diff_vec_analysis(
        self, y: np.ndarray, lines: pd.Series, delta: np.uint64
    ) -> pd.Series:
        diff_vec = np.abs(np.subtract(y, lines, dtype=np.int64))
        missing_val = np.where(diff_vec > delta / 20)[0]
        while missing_val.shape[0] > 0:
            if np.abs(diff_vec[missing_val[0]] - delta) / delta < 0.1:  # double line
                lines = np.concatenate(
                    (lines[: missing_val[0]], lines[missing_val[0] + 1 :])
                )
            else:
                lines = np.concatenate(
                    (
                        lines[: missing_val[0]],
                        np.atleast_1d(y[missing_val[0]]),
                        lines[missing_val[0] :],
                    )
                )
            # Restart the loop
            y = self.__gen_line_model(pd.Series(lines), delta)
            diff_vec = np.abs(np.subtract(y, lines, dtype=np.int64))
            missing_val = np.where(diff_vec > delta / 20)[0]
        return pd.Series(lines)

    def __finalize_lines(self, lines, delta) -> pd.Series:
        """ Sample the lines so that they're "silent" between frames """
        lines_between_frames = int(np.rint(self.frame_delay / delta))
        start_of_frame_idx = np.arange(
            start=0,
            stop=len(lines),
            step=self.num_of_lines + lines_between_frames,
            dtype=np.uint64,
        )
        end_of_frame_idx = start_of_frame_idx + self.num_of_lines
        exact_lines = [
            lines[slice(start, end)]
            for start, end in zip(start_of_frame_idx, end_of_frame_idx)
        ]
        exact_lines = np.array(list(chain.from_iterable(exact_lines)), dtype=np.uint64)
        self.sig_val.num_of_frames = len(end_of_frame_idx)

        return pd.Series(exact_lines, dtype=np.uint64)



from itertools import tee
import logging
from typing import Dict, Tuple

import pandas as pd
import numpy as np
import attr
import numba


@attr.s(slots=True)
class ScanImageLineValidator:
    sig_val = attr.ib()  # SignalValidator

    def __getattr__(self, item):
        return getattr(self.sig_val, item)

    def run(self) -> Tuple[Dict, np.uint64]:
        """
        Interpolate SI-specific line signals.

        Returns a dictionary containing the data and the mean difference between subsequent lines
        """
        lines = self.dict_of_data["Lines"].loc[:, "abs_time"].copy()
        (
            lines_mat,
            rel_idx,
            end_of_frames_idx,
            last_idx_of_row,
            rel_idx_non_end_frame,
        ) = self.__calc_line_parameters(lines=lines)
        if lines_mat.shape[0] == 1:  # single frame
            lines = pd.Series(lines_mat[0, : self.num_of_lines], dtype=np.uint64)
            delta = np.uint64(lines.diff().mean())
        else:
            theo_lines, delta = self.__gen_line_model(
                lines=lines, rel_idx=rel_idx, end_of_frames_idx=end_of_frames_idx
            )
            lines_mat = self.__diff_mat_analysis(
                y=theo_lines,
                lines_mat=lines_mat,
                last_idx=last_idx_of_row,
                delta=delta,
                rel_idx=rel_idx_non_end_frame,
            )
            lines = self.__finalize_lines(lines_mat=lines_mat)
        if self.bidir:
            lines = self.sig_val.add_phase_to_bidir_lines(lines=lines)
        self.dict_of_data["Lines"] = pd.DataFrame(
            lines, dtype=np.uint64, columns=["abs_time"]
        )
        return self.dict_of_data, delta

    def __gen_line_model(
        self, lines: pd.Series, rel_idx: np.ndarray, end_of_frames_idx: np.ndarray
    ):
        """ Using linear approximation generate a model for the "correct" line signal """
        rel_idx_first_frame = rel_idx[rel_idx < self.num_of_lines]
        lines_first_frame = lines.iloc[: self.num_of_lines].drop(rel_idx_first_frame)
        # Create a model: y = mx + const
        const = lines_first_frame[0]
        m = np.uint64(lines_first_frame.diff().median())
        x = np.arange(start=0, stop=self.num_of_lines, dtype=np.uint64)
        y = m * x + const
        y = np.tile(y, (end_of_frames_idx.shape[0], 1))
        y[1:, :] -= y[1, 0]
        y = np.concatenate((y, np.zeros_like(y)), axis=1)
        return y, m

    def __diff_mat_analysis(
        self,
        y: np.ndarray,
        lines_mat: np.ndarray,
        last_idx: np.ndarray,
        delta: int,
        rel_idx: np.ndarray,
    ) -> np.ndarray:
        """
        Check for missing/extra lines in the matrix of lines.

        :param np.ndarray diff_mat:
        :param np.ndarray last_idx: Last index of relevant lines in the frame
        """
        diff_mat = np.abs(np.subtract(lines_mat, y, dtype=np.int64))
        missing_vals_rows, missing_vals_cols = np.where(diff_mat > delta / 15)
        frames_to_correct = np.where(last_idx != self.num_of_lines)[0]
        for frame_num in frames_to_correct:
            num_of_missing_lines = last_idx[frame_num] - self.num_of_lines
            if num_of_missing_lines < 0:
                for miss in missing_vals_cols[missing_vals_rows == frame_num]:
                    lines_mat[frame_num, :] = np.concatenate(
                        (
                            lines_mat[frame_num, :miss],
                            np.atleast_1d(y[frame_num, miss]),
                            lines_mat[frame_num, miss:-1],
                        )
                    )
            elif num_of_missing_lines > 0:
                cur_missing_cols = missing_vals_cols[
                    missing_vals_rows == frame_num
                ].copy()
                iters = 0
                while (cur_missing_cols.shape[0] > 1) and (iters < 1000):
                    lines_mat[frame_num, :] = np.concatenate(
                        (
                            lines_mat[frame_num, : cur_missing_cols[0]],
                            np.atleast_1d(y[frame_num, cur_missing_cols[0]]),
                            lines_mat[frame_num, cur_missing_cols[1] :],
                        )
                    )
                    diff_line = np.abs(
                        np.subtract(
                            lines_mat[frame_num, :], y[frame_num, :], dtype=np.int64
                        )
                    )
                    cur_missing_cols = np.where(diff_line > delta / 20)[0]
                    iters += 1
                if iters == 1000:
                    logging.warning(
                        "Line signal was corrupt during at least one frame."
                    )
        return lines_mat

    def __finalize_lines(self, lines_mat: np.ndarray) -> pd.Series:
        """ Add back the start-of-frame times to the lines """
        frames_mat = np.array(self.dict_of_data["Frames"]).reshape(
            (len(self.dict_of_data["Frames"]), 1)
        )
        frames_mat = np.tile(frames_mat, (1, self.num_of_lines))
        lines_mat = lines_mat[:, : self.num_of_lines] + frames_mat
        return pd.Series(lines_mat.ravel())

    def __calc_line_parameters(
        self, lines: pd.Series
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """ Generate general parameters of the given acquisition """
        change = np.abs(lines.diff().pct_change(periods=1))
        rel_idx = np.where(change > self.change_thresh)[0]
        div, mod = divmod(len(lines), self.num_of_lines)
        if len(rel_idx) == 0 and mod != 0 and div == 0:
            raise UserWarning(
                "Data contained incomplete signals only from the first frame."
            )

        end_of_frames = change > 5
        end_of_frames_idx = np.where(end_of_frames)[0]  # scanimage specific
        rel_idx_non_end_frame = np.where(change <= self.change_thresh)[0]
        lines_mat, last_idx_of_row = iter_end_of_frames(
            lines.to_numpy(), end_of_frames_idx, self.num_of_lines
        )
        start_time_of_frames = lines[end_of_frames_idx].values
        start_time_of_frames = np.concatenate(
            (np.array([0], dtype=np.uint64), start_time_of_frames[:-1])
        ).reshape((end_of_frames_idx.shape[0], 1))
        lines_mat -= start_time_of_frames
        if "Frames" not in self.dict_of_data:
            self.dict_of_data["Frames"] = pd.DataFrame(
                start_time_of_frames.ravel(), columns=["abs_time"]
            )

        for row, last_idx in enumerate(last_idx_of_row):
            lines_mat[row, last_idx:] = 0

        return (
            lines_mat,
            rel_idx,
            end_of_frames_idx,
            last_idx_of_row,
            rel_idx_non_end_frame,
        )

    @staticmethod
    def __pairwise(iterable):
        """From itertools: s -> (s0,s1), (s1,s2), (s2, s3), ..."""
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)


@numba.jit(nopython=True)
def iter_end_of_frames(lines, end_of_frames_idx, num_of_lines):
    """Iterates over two signals - the start and end of each frame - and returns a
    matrix containing the lines between those two signals.
    """
    lines_mat = np.zeros((len(end_of_frames_idx), num_of_lines * 2), dtype=np.uint64)
    last_idx_of_row = np.zeros((end_of_frames_idx.shape[0]), dtype=np.int32)
    frame_start = 0
    for row, end in enumerate(end_of_frames_idx):
        frame_end = end
        last_idx = frame_end - frame_start
        lines_mat[row, :last_idx] = lines[frame_start:frame_end]
        last_idx_of_row[row] = last_idx
        frame_start = end
    return lines_mat, last_idx_of_row



import pytest
import pandas as pd
import numpy as np

from pysight.post_processing.flim import *


@pytest.fixture
def generate_flim():
    def _generate_flim(shape, flim, di=None, photons=None):
        if di is None:
            di = {1: np.array([1, 2, 3])}
        if photons is None:
            photons = pd.DataFrame(
                {"time_rel_pulse": np.arange(125), "Frames": 1, "Channel": 1}
            ).set_index(["Channel", "Frames"])
        summed_mem = di
        stack = di
        channels = pd.CategoricalIndex([1])
        out = PySightOutput(photons, summed_mem, stack, channels, shape, flim, dict())
        return LifetimeCalc(out, 1)

    return _generate_flim


@pytest.mark.skip
class TestFlim:
    shape = (10, 10, 10, 10)
    flim = True

    def test_get_photons(self, generate_flim):
        np.array_equal(
            generate_flim(self.shape, self.flim)._get_photons(), np.arange(125)
        )

    def test_aligned_hist(self, generate_flim):
        photons = np.arange(31)
        photons = np.concatenate((photons, [30, 30, 30, 29, 29, 28]))
        photons_df = pd.DataFrame(
            {"time_rel_pulse": photons, "Frames": 1, "Channel": 1}
        ).set_index(["Channel", "Frames"])
        hist = generate_flim(
            self.shape, self.flim, photons=photons_df
        )._gen_aligned_hist(photons)
        assert hist.argmax() == 0

    def test_lowest_bin(self, generate_flim):
        photons = np.linspace(0, 31, num=5, endpoint=False)
        photons = np.concatenate((photons, [30, 30, 30, 29, 29, 28]))
        photons = np.delete(photons, 0)
        photons_df = pd.DataFrame(
            {"time_rel_pulse": photons, "Frames": 1, "Channel": 1}
        ).set_index(["Channel", "Frames"])
        flim_obj = generate_flim(self.shape, self.flim, photons=photons_df)
        hist = flim_obj._gen_aligned_hist(photons)
        assert 2 == flim_obj._get_low_bin_idx(hist)



import attr
from attr.validators import instance_of
import pandas as pd
import numpy as np
import itertools
from typing import List, Generator
import psutil


@attr.s(slots=True)
class VolumeGenerator:
    """
    Generate the list of volume chunks to be processed.
    Main method is "create_frame_slices", which returns a generator containing
    slice objects that signify the chunks of volumes to be processed simultaneously.

    :param pd.DataFrame frames: Frames for the entire dataset. Should not contain a closing, right-edge, frame.
    :param tuple data_shape: Shape of the final n-dimensional array (from the Output object)
    :param int MAX_BYTES_ALLOWED: Number of bytes that can be held in RAM. Calculated using the
                                  ``psutil`` package if not supplied manually.
    """

    frames = attr.ib(validator=instance_of(pd.Series), repr=False)
    data_shape = attr.ib(validator=instance_of(tuple))
    MAX_BYTES_ALLOWED = attr.ib(default=0, validator=instance_of(int))
    num_of_frames = attr.ib(init=False)
    bytes_per_frames = attr.ib(init=False)
    full_frame_chunks = attr.ib(init=False)
    frame_slices = attr.ib(init=False)
    frames_per_chunk = attr.ib(init=False)
    num_of_chunks = attr.ib(init=False)

    def __attrs_post_init__(self):
        if self.MAX_BYTES_ALLOWED == 0:
            try:
                avail = psutil.virtual_memory().available
            except AttributeError:
                self.MAX_BYTES_ALLOWED = 1_000_000_000
            else:
                self.MAX_BYTES_ALLOWED = avail // 32  # magic number

    def create_frame_slices(self, create_slices=True) -> Generator:
        """
        Main method for the pipeline. Returns a generator with slices that
        signify the start time and end time of each chunk of frames. The indexing
        is inclusive-inclusive, and not inclusive-exclusive, since it's done using
        pandas' ``.loc`` method.

        :param bool create_slices: Used for testing, always keep ``True`` for actual code.
        """
        self.bytes_per_frames = np.prod(self.data_shape[1:]) * 8
        self.frames_per_chunk = int(
            max(1, self.MAX_BYTES_ALLOWED // self.bytes_per_frames)
        )
        self.num_of_frames = len(self.frames)
        self.num_of_chunks = int(max(1, len(self.frames) // self.frames_per_chunk))
        self.full_frame_chunks = self.__grouper()
        if create_slices:
            self.frame_slices = self.__generate_frame_slices()
            return self.frame_slices

    def __grouper(self) -> Generator:
        """ Chunk volume times into maximal-sized groups of values. """
        args = [iter(self.frames.values)] * self.frames_per_chunk
        return itertools.zip_longest(*args, fillvalue=np.nan)

    def __generate_frame_slices(self) -> Generator:
        if self.frames_per_chunk == 1:
            return (slice(frame[0], frame[0]) for frame in self.full_frame_chunks)

        start_and_end = []
        for chunk in self.full_frame_chunks:
            first, last = chunk[0], chunk[-1]
            if np.isnan(last):
                for val in reversed(chunk[:-1]):
                    if val is not np.nan:
                        last = val
                        break
            start_and_end.append((first, last))

        return (slice(np.uint64(t[0]), np.uint64(t[1])) for t in start_and_end)



import pandas as pd
import numpy as np
import attr
from typing import Dict, Tuple
import logging
from pysight.ascii_list_file_parser.apply_df_funcs import (
    get_lost_bit_np,
    get_lost_bit_tag,
    iter_string_hex_to_bin,
    convert_hex_to_int,
    convert_hex_to_bin,
)
from attr.validators import instance_of


@attr.s(slots=True)
class Tabulate(object):
    """
    Place all data in a pandas' DataFrame after sorting out each bit's meaning

    :param dict dict_of_inputs: Mapping of inputs to signal type
    :param np.ndarray data: Raw data from list file
    :param dict dict_of_slices_hex: The role of each bit in the data per timepatch used
    :parame int data_rang: Length in bins of a single sweep
    :param bool is_binary: Whether the list file is a binary one
    :param bool use_tag_bits: Are the TAG bits needed for image generation in this timepatch
    :param int time_after_sweep: How long does the multiscaler wait after a sweep
    :param int acq_delay: How long does the multiscaler wait before start of experiment
    :param int num_of_channel: Number of active channels
    """

    dict_of_inputs = attr.ib(validator=instance_of(dict))
    data = attr.ib(validator=instance_of(np.ndarray))
    dict_of_slices_hex = attr.ib(validator=instance_of(dict))
    data_range = attr.ib(default=1, validator=instance_of(int))
    use_tag_bits = attr.ib(default=False, validator=instance_of(bool))
    bitshift = attr.ib(default=0, validator=instance_of(int))
    time_after_sweep = attr.ib(default=int(96), validator=instance_of(int))
    acq_delay = attr.ib(default=int(0), validator=instance_of(int))
    num_of_channels = attr.ib(default=3, validator=instance_of(int))
    edge = attr.ib(init=False)
    channel = attr.ib(init=False)
    list_of_losts = attr.ib(init=False)
    df_after_timepatch = attr.ib(init=False)

    def run(self):
        """ Pipeline of analysis """
        self.__preparations_hex()
        self.__tabulate_input_hex()
        self.df_after_timepatch = self.__reformat_data_hex()
        self.__check_user_inputs()
        logging.info(
            "Sorted dataframe created. Starting to set the proper data channel distribution..."
        )

    @staticmethod
    def hex_to_bin_dict() -> Dict:
        """
        Create a simple dictionary that maps a hex input into a 4 letter binary output.
        """
        diction = {
            "0": "0000",
            "1": "0001",
            "2": "0010",
            "3": "0011",
            "4": "0100",
            "5": "0101",
            "6": "0110",
            "7": "0111",
            "8": "1000",
            "9": "1001",
            "a": "1010",
            "b": "1011",
            "c": "1100",
            "d": "1101",
            "e": "1110",
            "f": "1111",
        }
        return diction

    def __process_chan_edge(self, struct_of_data) -> Tuple[np.ndarray, np.ndarray]:
        """
        Simple processing scheme for the channel and edge data.
        """
        bin_array = np.array(iter_string_hex_to_bin("".join(struct_of_data.data)))
        edge = slice_string_arrays(bin_array, start=0, end=1)
        channel = slice_string_arrays(bin_array, start=1, end=4)

        return edge, channel

    def __preparations_hex(self) -> None:
        """
        "Set up the stage" for the tabulation process
        """
        if len(self.data) == 0:
            raise IOError("List file contained zero events.")

        for key in list(self.dict_of_slices_hex.keys())[1:]:
            self.dict_of_slices_hex[key].data = slice_string_arrays(
                self.data,
                self.dict_of_slices_hex[key].start,
                self.dict_of_slices_hex[key].end,
            )
        if not self.use_tag_bits:
            self.dict_of_slices_hex.pop("tag", None)

    def __tabulate_input_hex(self) -> None:
        """ Reformat the read hex data into a dataframe """
        # Channel and edge information
        self.edge, self.channel = self.__process_chan_edge(
            self.dict_of_slices_hex.pop("chan_edge")
        )
        # TODO: Timepatch == '3' is not supported because of this loop.

        if self.dict_of_slices_hex["lost"] is True:
            for key in list(self.dict_of_slices_hex.keys())[1:]:
                if self.dict_of_slices_hex[key].needs_bits:
                    list_with_lost = iter_string_hex_to_bin(
                        "".join(self.dict_of_slices_hex[key].data)
                    )
                    step_size = (
                        self.dict_of_slices_hex[key].end
                        - self.dict_of_slices_hex[key].start
                    )
                    if key == "tag":
                        (
                            self.list_of_losts,
                            self.dict_of_slices_hex[key].processed,
                        ) = get_lost_bit_tag(list_with_lost, step_size, len(self.data))
                    else:
                        (
                            self.list_of_losts,
                            self.dict_of_slices_hex[key].processed,
                        ) = get_lost_bit_np(list_with_lost, step_size, len(self.data))
                else:
                    if key == "tag":
                        self.dict_of_slices_hex[key].processed = convert_hex_to_bin(
                            self.dict_of_slices_hex[key].data
                        )
                    else:
                        self.dict_of_slices_hex[key].processed = convert_hex_to_int(
                            self.dict_of_slices_hex[key].data
                        )
        else:
            for key in list(self.dict_of_slices_hex.keys())[1:]:
                if key == "tag":
                    self.dict_of_slices_hex[key].processed = convert_hex_to_bin(
                        self.dict_of_slices_hex[key].data
                    )
                else:
                    self.dict_of_slices_hex[key].processed = convert_hex_to_int(
                        self.dict_of_slices_hex[key].data
                    )

    def __reformat_data_hex(self) -> pd.DataFrame:
        """
        Place the data in an organized DataFrame
        """
        df = pd.DataFrame(self.channel, columns=["channel"], dtype="category")
        df["edge"] = self.edge
        df["edge"] = df["edge"].astype("category")

        try:
            df["tag"] = self.dict_of_slices_hex["tag"].processed
        except KeyError:
            pass

        try:
            df[
                "lost"
            ] = self.list_of_losts  # TODO: Currently the LOST bit is meaningless
            df["lost"] = df["lost"].astype("category")
        except AttributeError:  # no data lost bit recorded
            pass

        df["abs_time"] = np.uint64(0)
        df["sweep"] = np.uint64(0)
        df["time_rel_sweep"] = np.uint64(0)

        if "sweep" in self.dict_of_slices_hex:
            # Each event has an acquisition delay before the start of sweep time,
            # and has to be multiplied by the sweep number and the time-after-sweep delay
            df["abs_time"] = (
                self.dict_of_slices_hex["time_rel_sweep"].processed
                + (
                    (self.dict_of_slices_hex["sweep"].processed - 1)
                    * ((self.data_range * self.bitshift) + self.time_after_sweep)
                )
                + self.dict_of_slices_hex["sweep"].processed * self.acq_delay
            )
            df["sweep"] = self.dict_of_slices_hex["sweep"].processed - 1
            df["time_rel_sweep"] = self.dict_of_slices_hex["time_rel_sweep"].processed
        else:
            df["abs_time"] = self.dict_of_slices_hex["time_rel_sweep"].processed

        return df

    def __check_user_inputs(self):
        """
        Assert that the channels that the user believe were recorded are actually there.
        Before sorting all photons make sure that no input is missing from the user. If it's missing
        the code will ignore this channel, but not raise an exception
        """

        actual_data_channels = set(
            self.df_after_timepatch["channel"].cat.categories.values
        )
        if actual_data_channels != set(self.dict_of_inputs.values()):
            logging.warning(
                "Channels that were inserted in GUI don't match actual data channels recorded. \n"
                f"The list files contains data in the following channels: "
                f"{{int(val, 2) for val in actual_data_channels}}."
            )
            thrown_channels = 0
            keys_to_pop = []
            for key, item in self.dict_of_inputs.items():
                if item not in actual_data_channels:
                    keys_to_pop.append(key)
                    thrown_channels += 1
            self.num_of_channels -= thrown_channels
            [self.dict_of_inputs.pop(key) for key in keys_to_pop]
        assert np.all(self.df_after_timepatch["abs_time"].values >= 0)  # safety check


def slice_string_arrays(arr: np.array, start: int, end: int) -> np.array:
    """
    Slice an array of strings efficiently.
    Based on http://stackoverflow.com/questions/39042214/how-can-i-slice-each-element-of-a-numpy-array-of-strings
    with modifications for Python 3.
    """
    b = arr.view("U1").reshape(len(arr), -1)[:, start:end]
    return np.frombuffer(b.tostring(), dtype="U" + str(end - start))



import logging
import gc
from typing import List, Tuple, Dict, Callable
from enum import Enum

import attr
from attr.validators import instance_of
import numpy as np
import pandas as pd
import zarr
from tqdm import tqdm

from .frame_chunk import FrameChunk, FlimCalc, HistWithIndex
from pysight.nd_hist_generator.outputs import DataShape, trunc_end_of_file


class ImagingSoftware(Enum):
    SCANIMAGE = "ScanImage"
    MSCAN = "MScan"


@attr.s
class Movie:
    """
    Creates FrameChunks that cn be appended into a multi-dimensional movie.

    To create the movie, call the `run()` method. The output will be according to the
    options specified in the GUI, although you can technically override them when
    instantiating this class.

    The Movie object also contains a `show_summed(channel)` method that can will show
    a two-dimensional projection of the multi-dimensional data.

    :param pd.DataFrame data: All recorded data
    :param pd.Series lines: Recorded lines
    :param Generator frame_slices: Slices of frame chunks
    :param pd.Series frames: Recorded frames
    :param float reprate: Laser repetition rate in Hz
    :param str name: Name of the file that is parsed
    :param float binwidth: Binwidth of multiscaler in seconds (100 ps == 100e-12)
    :param float fill_frac: Temporal fill fraction of the original movie, in percentage
    :param bool bidir: Whether the original scan was bi-directional
    :param pd.CategoricalIndex channels: Active and used channels in this iteration
    :param dict outputs: Required outputs
    :param bool censor: Whether to perform censor correction (currently NotImplemented)
    :param bool flim: Whether to perform FLIM
    :param dict lst_metadata: Metadata of the ``.lst`` file
    :param dict exp_params: Parameters used for fitting when using censor correction (currently NotImplemented)
    :param int line_delta: Number of bins between subsequent line signals
    :param float cache_size: Size of cache in bytes
    :param bool tag_as_phase: Whether to take into consideration the sinusoidal pattern of the TAG lens
    :param float tag_freq: Frequency of TAG lens in Hz
    :param float mirror_phase: Offset phase of the resonant mirror when scanning. If line shift (pixel shift) is
                                observed in the resulting image, change this parameter to fix it
    :param int num_of_frame_chunks: Number of chunks as calculated by the ``VolumeGenerator`` class
    :param int frames_per_chunk: Number of frames inside each chunk. ``frames_per_chunk * num_of_frame_chunks ==
                                len(frames)``
    :param tuple data_shape: Shape of data
    """

    data = attr.ib(validator=instance_of(pd.DataFrame), repr=False)
    lines = attr.ib(validator=instance_of(pd.Series), repr=False)
    frame_slices = attr.ib(repr=False)  # generator of frame slices from VolumeGenerator
    frames = attr.ib(validator=instance_of(pd.Series), repr=False)
    reprate = attr.ib(default=80e6, validator=instance_of(float))
    name = attr.ib(default="Movie", validator=instance_of(str))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    fill_frac = attr.ib(default=71.0, validator=instance_of(float))
    bidir = attr.ib(default=False, validator=instance_of(bool))
    channels = attr.ib(
        default=pd.CategoricalIndex([1]), validator=instance_of(pd.CategoricalIndex)
    )
    outputs = attr.ib(factory=dict, validator=instance_of(dict))
    censor = attr.ib(default=False, validator=instance_of(bool))
    flim = attr.ib(default=False, validator=instance_of(bool))
    lst_metadata = attr.ib(factory=dict, validator=instance_of(dict))
    line_delta = attr.ib(default=158_000, validator=instance_of(int))
    cache_size = attr.ib(default=10 * 1024 ** 3, validator=instance_of(int))
    tag_as_phase = attr.ib(default=True, validator=instance_of(bool))
    tag_freq = attr.ib(default=189e3, validator=instance_of(float))
    mirror_phase = attr.ib(default=-2.71, validator=instance_of(float))
    num_of_frame_chunks = attr.ib(default=1, validator=instance_of(int))
    frames_per_chunk = attr.ib(default=1, validator=instance_of(int))
    data_shape = attr.ib(default=(1, 512, 512), validator=instance_of(tuple))
    image_soft = attr.ib(
        default=ImagingSoftware.SCANIMAGE.value, validator=instance_of(str)
    )
    libver = attr.ib(default="latest", validator=instance_of(str))
    summed_mem = attr.ib(init=False, repr=False)
    stack = attr.ib(init=False, repr=False)
    x_pixels = attr.ib(init=False)
    y_pixels = attr.ib(init=False)
    z_pixels = attr.ib(init=False)
    flim_df = attr.ib(init=False)
    lifetime = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.x_pixels = self.data_shape[1]
        self.y_pixels = self.data_shape[2]
        self.z_pixels = 1
        if self.flim:
            self.flim_df = {i: list() for i in self.channels}
            self.lifetime = {i: 0 for i in self.channels}
        else:
            try:
                self.z_pixels = self.data_shape[3]
            except IndexError:
                pass
        self.line_delta = np.uint64(self.line_delta)

    def run(self) -> None:
        """
        Main pipeline for the movie object
        """
        funcs_during, funcs_end = self.__determine_outputs()
        self.__validate_df_indices()
        self.__process_data(funcs_during, funcs_end)
        self.__print_outputs()
        logging.info("Movie object created, analysis done.")

    def __determine_outputs(self) -> Tuple[List[Callable], List[Callable]]:
        """
        Based on the "outputs" variable, decide which outputs to generate.
        Returns a list of pointers to functions to execute for every volume, and after
        finishing generating the entire stack.
        """
        if not self.outputs:
            logging.warning(
                "No outputs requested. Data is still accessible using the dataframe variable."
            )
            return [], []

        funcs_to_execute_during = []
        funcs_to_execute_end = []

        if "memory" in self.outputs:
            self.summed_mem = {i: 0 for i in self.channels}
            self.stack = {i: list() for i in self.channels}
            funcs_to_execute_during.append(self.__create_memory_output)
            funcs_to_execute_end.append(self.__convert_list_to_arr)
            if "stack" in self.outputs:
                funcs_to_execute_end.append(self.__save_stack_at_once)
            if "summed" in self.outputs:
                funcs_to_execute_end.append(self.__save_summed_at_once)
            # TODO: Figure out what to do here
            # if "flim" in self.outputs:
            #     funcs_to_execute_end.append(self.__save_flim_at_once)

        else:
            if "stack" in self.outputs:
                self.outputs["stack"] = self.outputs["filename"].require_group(
                    "Full Stack"
                )
                funcs_to_execute_during.append(self.__save_stack_incr)

            if "summed" in self.outputs:
                self.summed_mem = {i: 0 for i in self.channels}
                funcs_to_execute_during.append(self.__append_summed_data)
                funcs_to_execute_end.append(self.__save_summed_at_once)

            if "flim" in self.outputs:
                funcs_to_execute_during.append(self.__append_flim_data)
                funcs_to_execute_end.append(self.__save_flim_at_once)
        return funcs_to_execute_during, funcs_to_execute_end

    def __process_data(
        self, funcs_during: List[Callable], funcs_end: List[Callable]
    ) -> None:
        """
        Create the outputs according to the outputs dictionary.
        Data is generated by appending to a list the needed micro-function to be executed.
        """
        # Execute the appended functions after generating each volume
        tq = tqdm(
            total=self.num_of_frame_chunks,
            desc=f"Processing frame chunks...",
            unit="chunk",
            leave=False,
        )
        for idx, frame_chunk in enumerate(self.frame_slices):
            sliced_df_dict, num_of_frames, frames, lines = self.__slice_df(frame_chunk)
            chunk = FrameChunk(
                movie=self,
                df_dict=sliced_df_dict,
                frames_per_chunk=num_of_frames,
                frames=frames,
                lines=lines,
            )
            hist_dict = chunk.create_hist()
            for func in funcs_during:
                for chan, hists in hist_dict.items():
                    func(data=hists[0], channel=chan, idx=idx, flim_hist=hists[1])

            tq.update(1)
            gc.collect()

        tq.close()
        [func() for func in funcs_end]

    def __slice_df(
        self, frame_chunk
    ) -> Tuple[Dict[int, pd.DataFrame], int, pd.Series, pd.Series]:
        """
        Receives a slice object and slices the DataFrame accordingly -
        once per channel. The returned dictionary has a key for each channel.
        """
        slice_dict = {}
        idx_slice = pd.IndexSlice
        for chan in self.data.index.levels[0].categories:
            slice_dict[chan] = self.data.loc[idx_slice[chan, frame_chunk], :]
        frames = self.frames.loc[frame_chunk]
        num_of_frames = len(frames)
        lines = self.lines.loc[frame_chunk]
        if len(lines) > self.x_pixels * num_of_frames:
            logging.warning(
                f"More-than-necessary line signals in the frame of chunk {frame_chunk}."
            )
        lines = lines.iloc[: self.x_pixels * num_of_frames]
        return slice_dict, num_of_frames, frames, lines

    def __validate_df_indices(self):
        """
        Make sure that the DataFrame of data contains the two
        important indices "Channel" and "Frames", and in the correct order.
        """
        if self.data.index.names[1] == "Lines":
            self.data = self.data.swaplevel()

        assert self.data.index.names[0] == "Channel"
        assert self.data.index.names[1] == "Frames"
        assert self.data.index.names[2] == "Lines"

    def __save_stack_at_once(self) -> None:
        """ Save the entire in-memory stack into a zarr file """
        z = self.outputs["filename"]
        logging.info("Saving full stack to disk...")
        for channel in self.channels:
            z["Full Stack"][f"Channel {channel}"][...] = self.stack[channel]

    def __save_summed_at_once(self) -> None:
        """ Save the entire in-memory summed data into a zarr file """
        for channel in self.channels:
            self.outputs["filename"]["Summed Stack"][f"Channel {channel}"][
                ...
            ] = np.squeeze(self.summed_mem[channel])

    def __convert_list_to_arr(self) -> None:
        """ Convert a list with a bunch of frames into a single numpy array with an extra
        dimension (0) containing the data. There will always be at least one
        frame of time data, this dimension is never squeezed out.
        """
        for channel in self.channels:
            new_stack = np.vstack(self.stack[channel])
            squeezed = np.squeeze(new_stack)
            if (
                new_stack.shape[0] != squeezed.shape[0]
            ):  # a single frame that was squeezed out
                self.stack[channel] = np.expand_dims(squeezed, axis=0)
            else:
                self.stack[channel] = squeezed

    def __create_memory_output(
        self, data: np.ndarray, channel: int, idx: int, flim_hist: np.ndarray
    ) -> None:
        """
        If the user desired, create two memory constructs -
        A summed array of all images (for a specific channel), and a stack containing
        all images in a serial manner.
        :param np.ndarray data: Data to be saved
        :param int channel: Current spectral channel of data
        :param int idx: Index of frame chunk
        :param np.ndarray flim_hist: FLIM data to be saved
        """
        self.stack[channel].append(data)
        assert len(data.shape) > 2
        self.summed_mem[channel] += np.uint16(data.sum(axis=0))

    def __save_stack_incr(
        self, data: np.ndarray, channel: int, idx: int, flim_hist: np.ndarray
    ) -> None:
        """
        Save incrementally new data to an open file on the disk
        :param np.ndarray data: Data to save
        :param int channel: Current spectral channel of data
        :param int idx: Index of frame chunk
        """
        cur_slice_start = self.frames_per_chunk * idx
        cur_slice_end = self.frames_per_chunk * (idx + 1)
        self.outputs["stack"][f"Channel {channel}"][
            cur_slice_start:cur_slice_end, ...
        ] = np.squeeze(data)

    def __append_summed_data(
        self, data: np.ndarray, channel: int, idx: int, flim_hist: np.ndarray
    ) -> None:
        """
        Create a summed variable later to be saved as the channel's data
        :param np.ndarray data: Data to be saved
        :param int channel: Spectral channel of data to be saved
        :param int idx: Index of frame chunk
        """
        assert len(data.shape) > 2
        self.summed_mem[channel] += np.uint16(data.sum(axis=0))

    def __append_flim_data(
        self, data: np.ndarray, channel: int, idx: int, flim_hist: pd.DataFrame
    ):
        flimcalc = FlimCalc(
            flim_hist["since_laser"].to_numpy(), flim_hist["bin"].to_numpy()
        )
        flimcalc.run()
        modified_data_shape = (self.frames_per_chunk,) + tuple(
            shape + 2 for shape in self.data_shape[1:]
        )
        modified_data_shape = DataShape(*modified_data_shape)
        flim_hist = flimcalc.histogram_result(modified_data_shape)
        flim_hist = np.nanmean(flim_hist, 0)
        print(np.nansum(flim_hist))
        self.flim_df[channel].append(np.nanmean(flim_hist, 0))

    def __save_flim_at_once(self):
        for chan in self.channels:
            mean_flim = np.nanmean(self.flim_df[chan], axis=0)
            np.save(f"data_chan_{chan}.npy", mean_flim)
            z = zarr.open(f'{self.outputs["filename"]}', "r+",)
            z["Lifetime"][f"Channel {chan}"][...] = np.squeeze(mean_flim)

    def __print_outputs(self) -> None:
        """ Print to console the outputs that were generated. """
        if not self.outputs:
            return

        logging.info(
            "======================================================= \nOutputs:\n--------"
        )
        if "stack" in self.outputs:
            logging.info(
                f'Stack file created with name "{self.outputs["filename"].store.path}", \ncontaining a data group named'
                ' "Full Stack", with one dataset per channel.'
            )

        if "memory" in self.outputs:
            logging.info(
                "The full data is present in dictionary form (key per channel) under `movie.stack`, "
                "and in stacked form under `movie.summed_mem`."
            )

        if "summed" in self.outputs:
            logging.info(
                f'Summed stack file created with name "{self.outputs["filename"].store.path}", \ncontaining a data group named'
                ' "Summed Stack", with one dataset per channel.'
            )

    def __nano_flim(self, data: np.ndarray) -> None:
        pass



from typing import Dict

import numpy as np
import pandas as pd


def add_bidir_lines(data: Dict[str, pd.DataFrame]) -> Dict[str, pd.DataFrame]:
    """
    For unidirectional scans fake line signals have to be inserted for us to identify forward- and
    back-phase photons.
    """

    length_of_lines = data["Lines"].shape[0]
    new_line_arr = np.zeros(length_of_lines * 2 - 1)
    new_line_arr[::2] = data["Lines"].loc[:, "abs_time"].values
    new_line_arr[1::2] = data["Lines"].loc[:, "abs_time"].rolling(window=2).mean()[1:]

    data["Lines"] = pd.DataFrame(new_line_arr, columns=["abs_time"], dtype="uint64")
    return data



"""
__author__ = Hagai Hargil
"""


import unittest
from pysight.nd_hist_generator.censor_correction import CensorCorrection
from pysight.nd_hist_generator.censor_correction import CensoredVolume
from pysight.nd_hist_generator.movie import Movie
import pandas as pd
from collections import namedtuple
import numpy as np


class TestCensorTools(unittest.TestCase):
    length = 100
    df = pd.DataFrame([i for i in range(length)], columns=["abs_time"])
    tuptype = namedtuple("TestCensor", ("hist", "edges"))
    cens = tuptype
    cens.hist = np.array([i for i in range(length)])
    cens.edges = np.array([1, 4, 8, 18, 25, length - 1])

    dict_of_data = {
        "PMT1": pd.DataFrame(
            [[1, 2, 3, 4, 5, 6, 7], [5, 6, 7, 8, 9, 10, 11]],
            columns=[
                "abs_time",
                "edge",
                "sweep",
                "time_rel_sweep",
                "time_rel_pulse",
                "bins_x",
                "bins_y",
            ],
            dtype=np.uint64,
        ),
        "Lines": pd.DataFrame(
            [[1, 2, 3, 4], [5, 6, 7, 8]],
            columns=["abs_time", "edge", "sweep", "time_rel_sweep"],
        ),
        "Laser": pd.DataFrame(
            [[0, 2, 3, 2], [3, 6, 7, 6]],
            columns=["abs_time", "edge", "sweep", "time_rel_sweep"],
        ),
    }
    dict_of_data["PMT1"].time_rel_pulse = dict_of_data["PMT1"].time_rel_pulse.astype(
        np.uint8
    )
    movie = Movie(
        data=dict_of_data["PMT1"],
        lines=dict_of_data["Lines"].abs_time,
        frames=pd.Series([]),
        frame_slices=(),
    )

    # def test_allocate_some_photons(self):
    #     censored = CensoredVolume(df=self.df, vol=Volume(data=self.dict_of_data['PMT1'],
    #                                                      lines=self.dict_of_data['Lines'].abs_time),
    #                               offset=0)
    #     photons = self.dict_of_data['PMT1']
    #     photons.set_index(keys=['bins_x', 'bins_y'], drop=True, inplace=True)
    #     idx_list = np.array([7])
    #     result = np.array([0], dtype=object)
    #     res_hist = np.histogram(np.array([5], dtype=object),
    #                             bins=np.arange(0, censored.bins_bet_pulses+1, dtype=np.uint8))[0]
    #     result[0] = res_hist
    #     returned = censored._CensoredVolume__allocate_photons_to_bins(idx_list, photons)
    #     self.assertTrue(np.all(result[0] == returned[0]))

    def test_append_laser(self):
        censored = CensorCorrection(
            raw=self.dict_of_data,
            movie=self.movie,
            all_laser_pulses=self.dict_of_data["Laser"],
            data=pd.DataFrame(),
        )

        censored.append_laser_line()
        res = pd.DataFrame(
            [[0, 2, 3, 2], [3, 6, 7, 6], [19, 0, 7, 22]],
            columns=["abs_time", "edge", "sweep", "time_rel_sweep"],
        )
        self.assertTrue(censored.raw["Laser"].equals(res))
